#
# Copyright 2018 Canonical Ltd.  All rights reserved.
#

import json
import os
import shutil
import subprocess
import time
import yaml

from subprocess import CalledProcessError, check_output, check_call, STDOUT

from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.contrib.network.ip import get_host_ip
from charmhelpers.core import templating
from charmhelpers.core.hookenv import (charm_name, config,
                                       local_unit, log, relation_get, relation_id, is_leader,
                                       status_set, unit_get, relation_ids,
                                       relations, unit_private_ip, relation_set,
                                       application_version_set, leader_get, leader_set)
from charmhelpers.core.host import (service_restart, service, service_running,
                                    add_group, add_user_to_group)
from charmhelpers.core.unitdata import kv
from charms.layer import ams_socket
from charms.layer.anbox import (remove_user_from_group,
                                remove_group,
                                close_old_and_open_new_ports,
                                reload_systemd,
                                install_or_upgrade_anbox_package,
                                remove_anbox_package)
from charms.layer.pki import (create_cert_files_if_needed, read_cert, fingerprint,
                              cert_raw_content)
from charms.reactive import hook, when, when_any, when_not
from charms.reactive.flags import clear_flag, is_flag_set, set_flag
from charms.reactive.helpers import data_changed
from charms.reactive.relations import endpoint_from_flag

from .ua import load_image_server_resource_token

DEB_PACKAGE_NAME = 'anbox-cloud-ams'
CONFIG_FOLDER = '/var/snap/ams/common'
SERVER_CONFIG_FOLDER = '{}/server'.format(CONFIG_FOLDER)
CLIENT_CONFIG_FOLDER = '{}/client'.format(CONFIG_FOLDER)
REGISTRY_CONFIG_FOLDER = '{}/registry'.format(CONFIG_FOLDER)
AMS_REGISTRY_CERT = '{}/client.crt'.format(REGISTRY_CONFIG_FOLDER)
SERVICE = 'snap.ams.ams.service'
FIRST_PORT_NUMBER = 10000
MAX_PORT_NUMBER = (1 << 16) - 1

PASSWORD_LENGTH = 64

UBUNTU_USER = 'ubuntu'

LXD_CONFIG_FOLDER = '{}/lxd'.format(CONFIG_FOLDER)
LXD_CLIENT_CERT_PATH = '{}/client.crt'.format(LXD_CONFIG_FOLDER)
LXD_CLIENT_KEY_PATH = '{}/client.key'.format(LXD_CONFIG_FOLDER)

CLIENT_CERTIFICATE_KEY = 'client_certificate'
FINGERPRINT_KEY = 'fingerprint'

PROMETHEUS_METRICS_PATH = '/internal/1.0/metrics'

STOPPING = 'charm.finished.ams'

SERVICE_DROP_IN_FOLDER = '/etc/systemd/system/{}.d'.format(SERVICE)
SERVICE_DROP_IN_NAME = '10-ams-unix-socket-chown.conf'
SERVICE_DROP_IN_PATH = '{}/{}'.format(SERVICE_DROP_IN_FOLDER,
                                      SERVICE_DROP_IN_NAME)

IMAGE_SERVER_OLD_URL = 'https://images.anbox-cloud.io/releases/'
IMAGE_SERVER_NEW_URL = 'https://images.anbox-cloud.io/stable/'

LXD_UNIT_PREFIX = 'lxd'
NODES_STATE_KEY_NAME = 'known_nodes'


class CertificateException(Exception):
    pass


class HttpException(Exception):
    pass


class NotFoundException(Exception):
    pass


class InvalidParameterException(Exception):
    pass


class NodeState:
    def __init__(self):
        # Load current state from leader data. We will never write to this
        # as the hook may loose leadership mid execution and the data might
        # be written by another unit already. We will reload the nodes list
        # when we commit and add all news nodes if the don't exist yet
        self._nodes = self._load()

        self._added_nodes = []

    def _load(self):
        raw_nodes = leader_get(NODES_STATE_KEY_NAME)
        if not raw_nodes:
            return []
        return json.loads(raw_nodes)

    def _id_from_name(self, name):
        if not name.startswith(LXD_UNIT_PREFIX):
            raise Exception('Invalid LXD node name {}'.format(name))

        return int(name[len(LXD_UNIT_PREFIX):])

    def add(self, name):
        self._added_nodes.append(self._id_from_name(name))

    def has(self, name):
        id = self._id_from_name(name)
        return id in self._nodes or id in self._added_nodes

    def commit(self):
        nodes_to_store = self._load()
        nodes_to_store.extend(self._added_nodes)

        # As we might have lost leadership during the hook execution we need to
        # reload the state here and add missing nodes
        leader_set({NODES_STATE_KEY_NAME: json.dumps(nodes_to_store)})

    def reset(self):
        leader_set({NODES_STATE_KEY_NAME: '[]'})
        self._nodes = []
        self._added_nodes = []


@when('ua.ready')
@when_not('ams.installed')
def install():
    install_or_upgrade_anbox_package(DEB_PACKAGE_NAME)
    set_charm_version()
    setup_charm()
    set_flag('ams.installed')


@hook('upgrade-charm')
def setup_charm():
    """
    Adds ubuntu user to ams group, just in case it is not still added
    Deploys the charm
    """
    if not is_flag_set('ams.installed'):
        return

    install_or_upgrade_anbox_package(DEB_PACKAGE_NAME)

    set_charm_version()
    add_group(charm_name())
    add_user_to_group(UBUNTU_USER, charm_name())
    create_amc_alias()
    force_setup_ams_as_notify_service()
    deploy()

    # If we have an NRPE relation, force run of that as well
    if relation_ids('nrpe-exteral-master'):
        clear_flag('nagios.configured')


def set_charm_version():
    try:
        info = subprocess.run(['snap', 'info', 'ams'],
                              stdout=subprocess.PIPE)
    except subprocess.CalledProcessError as err:
        log('Failed to get snap info: {}'.format(err), level='ERROR')
    else:
        stdout = info.stdout.decode('utf8').splitlines()
        version = [line.split()[1] for line in stdout if 'installed' in line]
        if version:
            application_version_set(version[0])


@hook('stop')
def stop():
    """
    Remove the snap and unset previous stored values in unit local
    database
    """
    # We need this flag to prevent races when stopping the charm.
    # Snap removal clears ams.installed flag, but that happens
    # when snap has completed the uninstall and meanwhile, other
    # reactive handlers could be triggered
    set_flag(STOPPING)
    clear_flag('ams.ready')

    remove_anbox_package(DEB_PACKAGE_NAME)

    remove_drop_in_for_ams_service()
    remove_user_from_group(UBUNTU_USER, charm_name())
    remove_group(charm_name())
    kv().flush()


@when_any('config.changed.port',
          'config.changed.port_range',
          'config.changed.log_level',
          'config.changed.prometheus_target_port',
          'config.changed.prometheus_metrics_path',
          'config.changed.prometheus_tls_cert_path',
          'config.changed.prometheus_tls_key_path',
          'config.changed.prometheus_basic_auth_username',
          'config.changed.prometheus_basic_auth_password',
          'config.changed.prometheus_extra_labels',
          'config.changed.metrics_server',
          'config.changed.registry_mode',
          'config.changed.lxd_project')
def deploy_config_changed():
    """
    Only handling port or log_level config parameter to redeploy.
    Rest of them apply when adding a new lxd unit
    """
    deploy()


@when('ams.installed')
@when_not('ams.prepared')
def snap_installed():
    create_amc_alias()
    setup_ams_as_notify_service()
    set_flag('ams.prepared')


# We have to put this code into the hook and not using the etcd.available
# flag because when departed hook is executed there is still connection
# to etcd and we can purge any remaining lxd node.
# This purge solves this next situation:
# 1.- add ams-etcd relation
# 2.- add ams-lxd relation
# 3.- remove ams-etcd relation
# 4.- remove ams-lxd relation
# 5.- add ams-etcd relation
# 6.- `amc node list` shows the old list of nodes, when it should not
# show any, because ams-lxd relation is currently not set.
@hook('etcd-relation-departed')
def purge_etcd(*args):
    if not is_leader():
        return

    if not service_running(SERVICE):
        return

    nodes = list_service_lxd_nodes()
    if nodes is None:
        return

    for node in nodes:
        if 'name' in node:
            remove_lxd_node_from_service(node['name'])
        # Don't allow additional management over this relation.
        # Otherwise, etcd.available flag could trigger the
        # update handler and purged nodes would be added back
        clear_flag('etcd.available')

    NodeState().reset()


@when('ams.installed', 'etcd.available', 'etcd.tls.available')
@when_not('etcd.prepared')
def etcd_available(ams, etcd):
    if not deploy():
        return
    setup_config()
    set_flag('etcd.prepared')


@when('ams.installed')
@when_not('etcd.available')
def etcd_not_available():
    clear_flag('etcd.prepared')
    deploy()
    status_set('blocked', 'Waiting for etcd')


@when('ams.installed', 'endpoint.lxd.joined')
def endpoint_lxd_joined(lxd):
    if not lxd:
        log('LXD endpoint still not available. Skip')
        return

    create_cert_files_if_needed(LXD_CLIENT_CERT_PATH, LXD_CLIENT_KEY_PATH)
    if relation_id():
        cert = read_cert(LXD_CLIENT_CERT_PATH)
        if cert:
            lxd.publish_info(client_certificate=cert)


@when_not('endpoint.lxd.joined')
def endpoint_lxd_broken():
    if not is_leader():
        return

    NodeState().reset()


def process_cluster_changes(lxd):
    if not is_leader():
        log('Not the leader, not attempting to add LXD node')
        return

    if not lxd or not lxd.services():
        log('LXD relation data still not available. Skip')
        return

    nodes = get_available_lxd_nodes(lxd.services())
    if len(nodes) <= 0:
        log('Not found LXD node to add. Skip')
        return

    update_lxd_nodes_in_service(nodes, use_node_state=True)

    # As last step we publish the current list of nodes we have in the
    # cluster so that the LXD units can unblock themself
    clustered_nodes = []
    for n in list_service_lxd_nodes():
        clustered_nodes.append(n['name'])
    lxd.publish_info(clustered_nodes=clustered_nodes)


@when('ams.installed', 'etcd.available', 'lxd.changed')
def endpoint_lxd_changed(etcd, lxd):
    """
    If got a not null node, add it to the service.
    If this is the first node to be added, it provides the password
    for nodes added later.
    """
    process_cluster_changes(lxd)
    clear_flag('lxd.changed')


@when('ams.installed', 'etcd.available')
@when_not('endpoint.lxd.joined', 'endpoint.lxd-cluster.joined')
def endpoint_lxd_not_available():
    """If got a not null node, removes it from the service"""
    if not is_leader():
        return

    name = relation_get('node')
    if not name:
        return

    remove_lxd_node_from_service(name.strip('\"'))


@when('ams.installed', 'endpoint.lxd-cluster.joined')
def endpoint_lxd_integrator_joined():
    create_cert_files_if_needed(LXD_CLIENT_CERT_PATH, LXD_CLIENT_KEY_PATH)
    cert = read_cert(LXD_CLIENT_CERT_PATH)

    lxd = endpoint_from_flag('endpoint.lxd-cluster.joined')
    lxd.register_cert(cert)


@when('ams.installed', 'etcd.available', 'endpoint.lxd-cluster.available')
def endpoint_lxd_integrator_available():
    """
    If got a not null node, add it to the service.
    If this is the first node to be added, it provides the password
    for nodes added later.
    """
    if not is_leader():
        return

    lxd = endpoint_from_flag('endpoint.lxd-cluster.available')
    nodes = lxd.endpoints()
    if not nodes:
        return

    for node in nodes:
        node['managed'] = True

    update_lxd_nodes_in_service(nodes, use_node_state=False)


@when('ams.installed', 'etcd.available')
@when_not('endpoint.lxd-cluster.joined', 'endpoint.lxd.joined')
def endpoint_lxd_integrator_not_available():
    """If got a not null node, removes it from the service"""
    if not is_leader():
        return

    name = relation_get('node')
    if not name:
        return

    remove_lxd_node_from_service(name.strip('\"'))


@when('ams.installed', 'etcd.available', 'endpoint.rest-api.available')
def endpoint_rest_api_available(etcd, rest_api):
    if not service_running(SERVICE):
        return

    if not rest_api or not rest_api.services():
        return

    for services in rest_api.services():
        if CLIENT_CERTIFICATE_KEY in services.keys():
            client_certificate = services[CLIENT_CERTIFICATE_KEY]
            register_client_certificate(client_certificate)

    rest_api.publish_info(port=config('port'), node=local_unit().replace("/", ""))


@when('ams.installed', 'etcd.available')
@when_not('endpoint.rest-api.available')
def endpoint_rest_api_not_available():
    fingerprint = kv().get(FINGERPRINT_KEY, default=None)
    if fingerprint:
        unregister_client_certificate(fingerprint)


@when('prometheus.available')
def prometheus_available():
    if relation_id():
        port = int(config('prometheus_target_port'))
        if port == 0:
            raise Exception(
                'Prometheus relation joined but target port not configured')
        relation_set(
            relation_id(), {'port': port,
                            'metrics_path': PROMETHEUS_METRICS_PATH})


@when('endpoint.registry-publisher.joined', 'ams.ready')
@when_not('endpoint.aar-published.departed')
def aar_push_endpoint_joined():
    if not is_leader():
        return

    aar = endpoint_from_flag('endpoint.registry-publisher.joined')
    register_ams_to_aar(aar, 'publisher')


@when('endpoint.registry-client.joined', 'ams.ready')
@when_not('endpoint.registry-client.departed')
def aar_pull_endpoint_joined():
    if not is_leader():
        return

    aar = endpoint_from_flag('endpoint.registry-client.joined')
    register_ams_to_aar(aar, 'client')


def register_ams_to_aar(aar, mode):
    """
    Register an AMS certificate on AAR

    :param aar: the endpoint will be different depending on the type
    of relation. It can either be a pull or push relation, which determines if
    the AMS client will be a publisher or a consumer
    :param mode: the mode AMS wants to register as. This is used to prevent bad
    relations (client with publisher for instance). Should be either
    'publisher' or 'client'
    """
    cert = read_cert(AMS_REGISTRY_CERT)
    if not cert:
        raise CertificateException('Could not read client certificate')

    aar.new_ams_client(cert, mode)


@when('endpoint.registry-publisher.changed', 'ams.ready')
@when_not('endpoint.registry-publisher.departed')
def aar_push_endpoint_changed():
    if not is_leader():
        return

    aar = endpoint_from_flag('endpoint.registry-publisher.changed')
    if aar is not None:
        set_aar_config(aar, 'publisher')

    clear_flag('endpoint.registry-publisher.changed')


@when('endpoint.registry-client.changed', 'ams.ready')
@when_not('endpoint.registry-client.departed')
def aar_pull_endpoint_changed():
    if not is_leader():
        return

    aar = endpoint_from_flag('endpoint.registry-client.changed')
    if aar is not None:
        set_aar_config(aar, 'client')

    clear_flag('endpoint.registry-client.changed')


@when('endpoint.grafana.joined')
def register_grafana_dashboards():
    if not is_leader():
        return

    grafana = endpoint_from_flag('endpoint.grafana.joined')
    with open("templates/grafana-dashboard-anbox-cloud.json") as f:
        json_dashboard = json.loads(f.read())
        grafana.register_dashboard(name='anbox-cloud-dashboard',
                                   dashboard=json_dashboard)


def set_aar_config(aar_endpoint, mode):
    """Set up configuration for an AAR server"""
    registry = aar_endpoint.registry()
    if registry is None:
        return

    if 'certificate' not in registry:
        return

    try:
        wait_for_ams_ready()
    except Exception:
        log('Cannot run AMS service')
        return

    register_client_certificate(registry['certificate'])

    try:
        args = ['/snap/bin/ams.amc', 'config', 'set']
        check_output([*args, 'registry.url', 'https://{}:{}'.
                     format(registry['ip'], registry['port'])], stderr=STDOUT)
        check_output([*args, 'registry.fingerprint', registry['fingerprint']],
                     stderr=STDOUT)

        mode = 'push' if mode == 'publisher' else 'pull'
        mode_override = config('registry_mode')
        if mode_override is not None and len(mode_override) > 0:
            mode = mode_override

        check_output([*args, 'registry.mode', mode], stderr=STDOUT)
    except CalledProcessError as ex:
        log('Error setting AAR on AMS: {} - {}'.format(
            ex.returncode, str(ex.output, 'utf-8')
        ))
        raise ex


@when('endpoint.registry-client.departed', 'ams.installed')
def remove_arr_pull_client():
    if not is_leader():
        return

    aar = endpoint_from_flag('endpoint.registry-client.departed')
    aar.remove_ams_client()
    reset_aar_config()
    clear_flag('endpoint.registry-client.departed')


@when('endpoint.registry-publisher.departed', 'ams.installed')
def remove_arr_push_client():
    if not is_leader():
        return

    aar = endpoint_from_flag('endpoint.registry-publisher.departed')
    aar.remove_ams_client()
    reset_aar_config()
    clear_flag('endpoint.registry-publisher.departed')


def reset_aar_config():
    try:
        args = ['/snap/bin/ams.amc', 'config', 'set']
        check_output([*args, 'registry.url', ''], stderr=STDOUT)
        check_output([*args, 'registry.fingerprint', ''], stderr=STDOUT)
        check_output([*args, 'registry.mode', 'pull'], stderr=STDOUT)
    except CalledProcessError as ex:
        log('Error resetting AAR config on AMS: {} - {}'.format(
            ex.returncode,
            str(ex.output, 'utf-8')
        ))
        raise ex


def deploy_needed(etcd_config, metrics_server, ip, port, log_level,
                  prometheus_port, port_range, prometheus_tls_cert_path,
                  prometheus_tls_key_path, prometheus_basic_auth_username,
                  prometheus_basic_auth_password, lxd_project):
    """
    Returns True if any of received params has changed since last time
    or if service is not running
    """
    return data_changed('service_running', service_running(SERVICE))\
        or data_changed('etcd_config', etcd_config)\
        or data_changed('metrics_server', metrics_server)\
        or data_changed('ip', ip)\
        or data_changed('port', port)\
        or data_changed('log_level', log_level)\
        or data_changed('prometheus_target_port', prometheus_port)\
        or data_changed('port_range', port_range)\
        or data_changed('prometheus_tls_cert_path', prometheus_tls_cert_path)\
        or data_changed('prometheus_tls_key_path', prometheus_tls_key_path)\
        or data_changed('prometheus_basic_auth_username',
                        prometheus_basic_auth_username)\
        or data_changed('prometheus_basic_auth_password',
                        prometheus_basic_auth_password)\
        or data_changed('lxd_project', lxd_project)


def deploy():
    """
    Generates a configuration from templates. Writes out that configuration
    to a settings.yaml file, restarts the snap and updates listening port
    """
    if is_flag_set(STOPPING):
        return False

    etcd_config = get_etcd_config()
    if etcd_config is None:
        return False
    else:
        # We can also not deploy AMS when we have no etcd servers published on
        # the relation yet
        if ('connection_string' not in etcd_config or
                'client_ca' not in etcd_config or
                'client_key' not in etcd_config or
                'client_cert' not in etcd_config):
            return False
        etcd_servers = etcd_config["connection_string"].split(',')
        if len(etcd_servers) == 0:
            return False

    ip = get_host_ip(unit_get('private-address'))
    port = config('port')
    log_level = config('log_level')
    prometheus_port = config('prometheus_target_port')
    prometheus_tls_cert_path = config('prometheus_tls_cert_path')
    prometheus_tls_key_path = config('prometheus_tls_key_path')
    prometheus_basic_auth_username = config('prometheus_basic_auth_username')
    prometheus_basic_auth_password = config('prometheus_basic_auth_password')
    metrics_server = config('metrics_server')
    port_range = config('port_range')
    lxd_project = config('lxd_project')
    # Don't deploy if not found any new config information
    if not deploy_needed(etcd_config, metrics_server, ip, port, log_level,
                         prometheus_port, port_range, prometheus_tls_cert_path,
                         prometheus_tls_key_path,
                         prometheus_basic_auth_username,
                         prometheus_basic_auth_password,
                         lxd_project):
        status_set('active', '')
        return False

    if not is_flag_set('ams.installed'):
        status_set('active', '')
        return False

    log('Deploying...')
    status_set('maintenance', 'Deploying service')

    configure_service(etcd_config, metrics_server, ip, port, log_level,
                      prometheus_port, port_range, lxd_project)

    service('enable', SERVICE)
    service_restart(SERVICE)

    add_group(charm_name())

    add_user_to_group(UBUNTU_USER, charm_name())

    wait_for_ams_ready()

    set_unix_socket_group_ownership(charm_name())

    close_old_and_open_new_ports('port')

    set_flag('ams.ready')

    status_set('active', '')

    return True


@when('ams.ready', 'ua.attached', 'etcd.prepared')
@when_not('ams.image_server.ready')
def configure_image_server():
    # First we have to see if we need to migrate from the old image server to
    # the new one.
    output = check_output(['/snap/bin/amc', 'config', 'show']).strip()
    config = yaml.load(str(output, 'utf-8'))
    current_server_url = config['config']['images.url']
    need_resource_token = False
    if current_server_url == IMAGE_SERVER_OLD_URL:
        check_call(['/snap/bin/amc', 'config', 'set', 'images.url', IMAGE_SERVER_NEW_URL])
        need_resource_token = True
    elif current_server_url == IMAGE_SERVER_NEW_URL:
        need_resource_token = True

    # We only configure the resource token when we're configured with the right
    # server. We will check again the next time our handler is called (during
    # update-status most likely)
    if not need_resource_token:
        return

    resource_token = load_image_server_resource_token()
    auth_details = 'bearer:{}'.format(resource_token)
    check_call(['/snap/bin/amc', 'config', 'set', 'images.auth', auth_details])

    set_flag('ams.image_server.ready')


def create_amc_alias():
    """Create alias for ams.amc command"""
    if os.path.exists('/snap/bin/amc'):
        return

    try:
        check_output(['snap', 'alias', 'ams.amc', 'amc'], stderr=STDOUT)
    except CalledProcessError as ex:
        log('Error creating alias for ams.amc command: {} - {}'.format(
            ex.returncode, ex.output))
        raise ex


def set_unix_socket_group_ownership(grp):
    """
    Set unix socket group ownership
    Note: the unix socket is created on the fly during service startup. We
    won't change the unix socket group ownership until ams is up and running.
    """
    try:
        wait_for_ams_ready()
        ams_socket.change_group_ownership(grp)
    except Exception:
        log('Could not change ams unix socket group to {} \
             because AMS service is not running'.format(grp))
        return


def _write_to_file(path, content):
    parent = os.path.dirname(path)
    if not os.path.isdir(parent):
        os.makedirs(parent)
    with open(path, 'w') as stream:
        stream.write(content)


def configure_service(etcd_config, metrics_server, ip, port, log_level,
                      prometheus_port, port_range, lxd_project):
    """Creates service configuration based on received params."""
    if not os.path.exists(SERVER_CONFIG_FOLDER):
        os.mkdir(SERVER_CONFIG_FOLDER, mode=0o755)

    prometheus_enabled = False
    prometheus_ip = unit_private_ip()
    prometheus_extra_labels = []
    if prometheus_port > 0:
        prometheus_enabled = True
        close_old_and_open_new_ports('prometheus_target_port')
        prometheus_tls_key_path = config('prometheus_tls_key_path') or ""
        prometheus_tls_cert_path = config('prometheus_tls_cert_path') or ""
        prometheus_username = config('prometheus_basic_auth_username') or ""
        prometheus_password = config('prometheus_basic_auth_password') or ""

        labels = config('prometheus_extra_labels').split(",")
        prometheus_extra_labels = {}
        for label in labels:
            parts = label.split("=")
            if len(parts) != 2:
                log("Invalid extra label option: {}".format(label))
                continue
            prometheus_extra_labels[parts[0]] = parts[1]

    validate_port_range(port_range)

    etcd_servers = ''
    etcd_key = ''
    etcd_cert = ''
    etcd_ca = ''
    if etcd_config:
        certs_path = '/var/snap/ams/common/etcd'
        if ('client_ca' in etcd_config and 'client_key' in etcd_config and
                'client_cert' in etcd_config):
            etcd_ca = "{}/client-ca.pem".format(certs_path)
            etcd_cert = "{}/client-cert.pem".format(certs_path)
            etcd_key = "{}/client-key.pem".format(certs_path)

            _write_to_file(etcd_ca, etcd_config['client_ca'])
            _write_to_file(etcd_cert, etcd_config['client_cert'])
            _write_to_file(etcd_key, etcd_config['client_key'])

        if 'connection_string' in etcd_config:
            etcd_servers = etcd_config["connection_string"].split(',')

    if metrics_server:
        # Anbox needs influxdb: prefix for the metrics server. Otherwise
        # containers won't send any data to telegraf socket listener
        metrics_server = 'influxdb:{}'.format(metrics_server)

    templating.render(
        source='settings.yaml',
        target='{}/settings.yaml'.format(SERVER_CONFIG_FOLDER),
        context={
            'ip': ip,
            'port': port,
            'log_level': log_level,
            'port_range': port_range,
            'etcd_use_embedded': len(etcd_servers) == 0,
            'etcd_servers': etcd_servers,
            'etcd_key': etcd_key,
            'etcd_cert': etcd_cert,
            'etcd_ca': etcd_ca,
            'metrics_server': metrics_server,
            'prometheus_enabled': prometheus_enabled,
            'prometheus_ip': prometheus_ip,
            'prometheus_port': prometheus_port,
            'prometheus_tls_key': prometheus_tls_key_path,
            'prometheus_tls_cert': prometheus_tls_cert_path,
            'prometheus_basic_auth_username': prometheus_username,
            'prometheus_basic_auth_password': prometheus_password,
            'prometheus_extra_labels': prometheus_extra_labels,
            'lxd_project': lxd_project,
        },
        perms=0o600)


def wait_for_ams_ready():
    """
    Repeats until AMS service is ready to use
    """
    do_until_completed(is_ams_ready)


def is_ams_ready():
    """
    Tests if systemd service is running. Once done, requests /1.0
    url to the local unix socket and verifies it works to return true
    """
    if not service_running(SERVICE):
        log('AMS service is not running')
        return False

    result_code, _ = ams_socket.get('1.0')
    log('AMS return result code {}'.format(result_code))
    return result_code == 200


def do_until_completed(fn, times=150, step=2):
    """
    Repeats the function n times with a step of m seconds until
    no exception is thrown. If function returns a boolean value, a
    True is also required.
    When max times is reached an exception is thrown.
    """
    for i in range(times):
        try:
            output = fn()
            if isinstance(output, bool) and not output:
                log('{} retry number {}'.format(fn.__name__, i+1))
            else:
                return output
        except Exception as ex:
            log('Error executing {} : {}. Retry {}'.format(
                fn.__name__, ex, i+1))
        time.sleep(step)

    raise Exception('Timeout executing {}'.format(fn.__name__))


def update_lxd_nodes_in_service(nodes, use_node_state):
    """
    Updates existing AMS nodes with received ones, by adding
    and removing as needed
    """
    if not service_running(SERVICE):
        log('AMS service is not running, cannot update LXD nodes')
        return

    current_nodes = list_service_lxd_nodes()
    nodes_to_add = nodes
    nodes_to_remove = []

    if current_nodes is not None:
        for current in current_nodes:
            found = False
            for remaining in nodes_to_add:
                if current['name'] == remaining['name']:
                    nodes_to_add.remove(remaining)
                    found = True
                    break

            if not found:
                nodes_to_remove.append(current['name'])

    if use_node_state:
        state = NodeState()

        # If we have nodes in our list which are already recorded as being added before
        # (independent of what check with AMS gave as result) we need to exclude them
        # here.
        cleaned_nodes_to_add = []
        for n in nodes_to_add:
            name = n['name']
            if state.has(name):
                log('Node {} is already added, skipping'.format(name), level='INFO')
                continue
            cleaned_nodes_to_add.append(n)

        nodes_to_add = cleaned_nodes_to_add

    if len(nodes_to_add) > 0:
        log("We have {} LXD nodes to add to AMS".format(len(nodes_to_add)))
    if len(nodes_to_remove) > 0:
        log("We have {} LXD nodes to remove from AMS".format(len(nodes_to_remove)))

    try:
        for n in nodes_to_add:
            if add_lxd_node_to_service(n) and use_node_state:
                state.add(n['name'])
    finally:
        if use_node_state:
            # We have to commit the state with all nodes which got added
            # regardless if some of the nodes failed as they are already
            # added to AMS and part of the LXD cluster
            state.commit()

    for m in nodes_to_remove:
        remove_lxd_node_from_service(m)


def add_lxd_node_to_service(node):
    """
    Adds a new LXD node to the service.
    Requires the service to be running.
    """
    if not node:
        log('Cannot add invalid node to AMS')
        return False
    if 'name' not in node:
        log('Cannot add node without name to AMS')
        return False
    if 'cert_fingerprints' not in node:
        log('Cannot add node without certificate fingerprint to AMS')
        return False

    fps = node['cert_fingerprints']
    if not client_cert_is_registered_on_lxd(fps):
        log('AMS certficate is not registered with LXD')
        return False

    try:
        wait_for_ams_ready()
    except Exception:
        log('Could not add LXD node {} because AMS service is not running'.format(node['name']))
        return False

    if lxd_node_exists_in_service(node['name']):
        log('Not adding existing node to AMS')
        return True

    node_addr = node.get('endpoint')
    if node_addr:
        node_addr = node_addr.replace('https://', '').replace('http://', '').split(':')[0]
    else:
        node_addr = node['private_address']
    log('Adding LXD node {} at {}'.format(node['name'], node_addr))

    # A storage device set via the configuration always takes precedence over
    # one set from the LXD charm. If nothing is set we let AMS choose a good
    # default.
    storage_device = config('storage_device')
    if len(storage_device) == 0:
        if 'storage_device' in node.keys():
            storage_device = node['storage_device']

    storage_pool = config('storage_pool') or ""

    cmd = ['amc', 'node', 'add', node['name'], node_addr]

    lxd_managed = False
    if 'managed' in node and node['managed'] is True:
        lxd_managed = True
        cmd.extend(['--unmanaged'])

    # We can only use a storage pool if the LXD cluster is not managed by us
    if len(storage_pool) > 0 and lxd_managed:
        cmd.extend(['--storage-pool', storage_pool])
    elif len(storage_device) > 0:
        cmd.extend(['--storage-device', storage_device])

    if 'mtu' in node and int(node['mtu']) > 0:
        cmd.extend(['--network-bridge-mtu', str(node['mtu'])])

    # Give the command a timeout to not wait forever
    cmd.extend(['--timeout', '5m'])

    # At times LXD is restarting or not available for a short moment. If
    # that happens we will retry adding the node after a short timeout
    num_retries = 10
    for n in range(num_retries):
        try:
            check_output(cmd, stderr=STDOUT)
            break
        except CalledProcessError as ex:
            err = str(ex.output, 'utf-8')
            if "Unable to connect" in err and n < num_retries - 1:
                log('LXD not available, retrying')
                time.sleep(10)
                continue

            log('Error adding LXD node {} to AMS: {} - {}'.format(
                node['name'], ex.returncode, err))
            raise ex

    # Set additional node configuration after created
    if 'cpu_cores' in node and node['cpu_cores']:
        node_set(node['name'], 'cpus', node['cpu_cores'])
    if 'memory' in node and node['memory']:
        node_set(node['name'], 'memory', node['memory'])
    if 'public_address' in node and node['public_address']:
        node_set(node['name'], 'public-address', node['public_address'])

    log('LXD node {} added at {}'.format(node['name'], node_addr))

    return True


def node_set(node, key, value):
    cmd = ['amc', 'node', 'set', node, key, str(value)]

    try:
        wait_for_ams_ready()
    except Exception:
        log('Could not set LXD node property because AMS service \
            is not running')
        return

    try:
        check_output(cmd, stderr=STDOUT)
    except CalledProcessError as ex:
        log('Error setting {} node configuration {} = {}: {} - {}'.format(
            node, key, str(value), ex.returncode, str(ex.output, 'utf-8')))
        raise ex


def remove_lxd_node_from_service(name):
    """
    Removes from the service the LXD node whose name matches
    the received parameter.
    """
    if not name:
        return

    try:
        wait_for_ams_ready()
    except Exception:
        log('Could not remove LXD node {} because AMS service is not running'.
            format(name))
        return

    if not lxd_node_exists_in_service(name):
        return

    log('Removing node {}'.format(name))

    try:
        check_output([
            '/snap/bin/ams.amc', 'node', 'remove',
            '--yes', '--force', '--keep-in-cluster', name],
            stderr=STDOUT)
    except CalledProcessError as ex:
        log('Error removing LXD node {} from AMS: {} - {}'.format(
            name, ex.returncode, str(ex.output, 'utf-8')))
        raise ex

    log('LXD node {} removed'.format(name))


def lxd_node_exists_in_service(name):
    """
    Returns true if service has a node whose name matches
    the received parameter
    """
    if not name:
        return False

    try:
        wait_for_ams_ready()
    except Exception as ex:
        log('Could not check if node {} exist in service \
            because AMS service is not running'.format(name))

        # We MUST raise the exception here. Returning false is not an option as that
        # will let the logic in add_lxd_node_to_service think the node wasn't added yet
        # and will attempt to add it. If we fail here we MUST signal something bad
        # happened with AMS and we have to stop the current hook execution to prevent
        # state corruption
        raise ex

    try:
        check_output(['/snap/bin/ams.amc', 'node', 'show', name],
                     stderr=STDOUT)
    except CalledProcessError as ex:
        if 'not found' in str(ex.output):
            return False
        log('Failed to check if LXD node {} is registered with AMS: {} - {}'.
            format(name, ex.returncode, str(ex.output, 'utf-8')))

        # The same as in the comment above applies here!
        raise ex

    return True


def list_service_lxd_nodes():
    """Returns the list of LXD nodes currently in the service"""
    try:
        wait_for_ams_ready()
    except Exception:
        log('Could not list LXD nodes because AMS service is not running')
        return []

    try:
        output = check_output(
            ['/snap/bin/ams.amc', 'node', 'list', '--format', 'json'],
            stderr=STDOUT)
        data = json.loads(str(output, 'utf-8'))
        if not data:
            return []
        return data
    except CalledProcessError as ex:
        log('Error fetching current LXD nodes from AMS: {} - {}'.format(
            ex.returncode, str(ex.output, 'utf-8')))
        raise ex


def get_available_lxd_nodes(services):
    """
    Returns the list of LXD nodes available in the relation with the
    LXD units.
    """
    nodes = []
    for srv in services:
        ks = srv.keys()
        if 'node' in ks:
            node = {'name': srv['node']}
            if 'private_address' in ks:
                node['private_address'] = srv['private_address']
            if 'public_address' in ks:
                node['public_address'] = srv['public_address']
            if 'port' in ks:
                node['port'] = srv['port']
            if 'cpu_cores' in ks:
                node['cpu_cores'] = srv['cpu_cores']
            if 'memory' in ks:
                node['memory'] = srv['memory']
            if 'mtu' in ks:
                node['mtu'] = srv['mtu']
            if 'cert_fingerprints' in ks:
                node['cert_fingerprints'] = srv['cert_fingerprints']
            if 'storage_device' in ks:
                node['storage_device'] = srv['storage_device']

            nodes.append(node)

    return nodes


def get_etcd_config():
    """
    Obtains etcd configuration from available relation.
    Does nothing if the relation is not set
    """
    rel_ids = list(relation_ids('etcd'))
    if len(rel_ids) == 0:
        log("etcd relation not ready set... skipping it for now")
        return None

    etcd_config = None
    etcd_id = rel_ids[0]
    relations_items = relations()['etcd'][etcd_id].items()
    for key, value in relations_items:
        if key.startswith('etcd'):
            etcd_config = value

    return etcd_config


def validate_port_range(port_range):
    """
    Validates port range to be assigned for container creation
    """
    port_parts = port_range.split('-')
    if len(port_parts) != 2:
        raise InvalidParameterException('Invalid port range')

    for port in port_parts:
        try:
            value = int(port)
            if value < FIRST_PORT_NUMBER or value > MAX_PORT_NUMBER:
                raise InvalidParameterException("Port out of range")
        except ValueError as ex:
            raise InvalidParameterException(ex)

    if int(port_parts[0]) >= int(port_parts[1]):
        raise InvalidParameterException("Invalid port range")


def client_cert_is_registered_on_lxd(fingerprints):
    """
    Returns true if our client certificate is
    amongst registered ones in lxd
    """
    cert = read_cert(LXD_CLIENT_CERT_PATH)
    if not cert:
        raise CertificateException('Could not read client certificate')

    cert_fp = fingerprint(cert)
    if not cert_fp:
        raise CertificateException('Could not get certificate fingerprint')

    for fp in fingerprints:
        if fp == cert_fp:
            return True
    return False


def register_client_certificate(cert):
    """
    Registers a client certificate through AMS REST interface.
    Certificate fingerprint is stored in kv to be used when removing
    the certificate from the trusted list if the relation is broken
    """
    raw_content = cert_raw_content(cert)
    if not raw_content:
        raise CertificateException('Could not get certificate raw content')

    payload = {
        "type": "client",
        "certificate": raw_content
    }
    result_code, result_text = ams_socket.post(
        '1.0/certificates', json.dumps(payload))

    if 200 <= result_code < 300:
        log('Registered client certificate: {}'.format(cert))
        # Save cert fingerprint in kv
        cert_fingerprint = fingerprint(cert)
        if not cert_fingerprint:
            raise CertificateException('Could not get certificate fingerprint')
        kv().set(FINGERPRINT_KEY, cert_fingerprint)
        return

    if "already exists" in result_text:
        log('Certificate already provisioned. Skip provision')
    else:
        raise CertificateException(result_text)


def unregister_client_certificate(fingerprint):
    """Unregisters a client certificate through AMS REST interface"""
    if not fingerprint:
        return
    result_code, _ = ams_socket.delete(
        '1.0/certificates/{}'.format(fingerprint))
    if 200 <= result_code < 300:
        log('Removed client certificate with fingerprint: {}'
            .format(fingerprint))
        return
    log('Warn: Could not unregister certificate')


def force_setup_ams_as_notify_service():
    remove_drop_in_for_ams_service()
    setup_ams_as_notify_service()


def setup_ams_as_notify_service():
    if os.path.exists(SERVICE_DROP_IN_PATH):
        return

    check_call(['snap', 'connect', 'ams:daemon-notify', 'core:daemon-notify'])
    deploy_drop_in_for_ams_service()


def deploy_drop_in_for_ams_service():
    ''' Deploys a drop in for ams systemd service to chown
    unix socket to 'ams' group after any restart'''
    if not os.path.exists(SERVICE_DROP_IN_FOLDER):
        os.mkdir(SERVICE_DROP_IN_FOLDER, mode=755)

    templating.render(
        source=SERVICE_DROP_IN_NAME,
        target=SERVICE_DROP_IN_PATH,
        context={},
        perms=0o644)

    reload_systemd()
    service_restart(SERVICE)


def remove_drop_in_for_ams_service():
    '''Removes the ams systemd drop-in for having
    ams as a systemd notify service'''
    if not os.path.exists(SERVICE_DROP_IN_FOLDER):
        return

    shutil.rmtree(SERVICE_DROP_IN_FOLDER)


@when('ams.installed',
      'etcd.available',
      'config.changed.config')
def config_changed():
    setup_config()


def setup_config():
    if not is_ams_ready():
        return

    cfg = config('config')
    if cfg is None or len(cfg) == 0:
        return

    options = cfg.split('\n')
    for option in options:
        parts = option.split("=")
        if len(parts) != 2:
            log("Invalid configuration option: {}".format(option))
            continue

        name = parts[0]
        value = parts[1]

        try:
            check_output([
                '/snap/bin/ams.amc', 'config', 'set', name, value],
                stderr=STDOUT)
        except CalledProcessError as ex:
            log('Error setting `{}` to `{}`'.format(name, value))
            raise ex


@when_any('config.changed.nagios_context',
          'config.changed.nagios_servicegroups')
def nagios_config_changed():
    if not is_flag_set('endpoint.nrpe-external-master.joined'):
        return
    update_nrpe_config()


@when('nrpe-external-master.available')
@when_not('nagios.configured')
def update_nrpe_config():
    hostname = nrpe.get_nagios_hostname()
    address = get_host_ip(unit_get('private-address'))
    port = config('port')
    nrpe_setup = nrpe.NRPE(hostname=hostname)
    cmd = [
        '/usr/lib/nagios/plugins/check_http',
        '--ssl', '--hostname', "{}:{}".format(address, port),
        '--url', '/1.0', '--expect', '200',
        '--warning', '5', '--critical', '10'
    ]
    nrpe_setup.add_check(
        shortname='ams_https',
        description='Check AMS https endpoint',
        check_cmd=" ".join(cmd),
    )
    nrpe_setup.write()
    set_flag('nagios.configured')
